//------------------------------------------------------------------------------
/*
 * Author:: Tod Gentille
 * Date::  Nov 18, 2009
 * Workfile:: tcpServer.cpp
 * Revision:: 1
 * (C) Copyright Syncor Systems, Inc.
 */
//------------------------------------------------------------------------------


#include <startnet.h>
#include <constants.h>
#include <tcp.h>
#include "ShellCommon.h"
#include "TcpServer.h"
#include "ServiceProvider.h"
#include <iostream>

using namespace std;


OS_MBOX TcpServer::_mailbox;
OS_SEM TcpServer::_msgProcessed;


TcpServer::TcpServer(TcpServerReset* theTcpServerReset) :
	_messageWaiting(false), _tcpServerReset(theTcpServerReset), _connectionIsOpen(false), _debuggingTcp(true)
{


}

OS_MBOX& TcpServer::GetMailbox()
{
	return _mailbox;
}

OS_SEM& TcpServer::GetSemaphore()
{
	return _msgProcessed;
}

void TcpServer::InitializeWrapper(void* const pd)
{
	ServiceProvider::GetTcpServerAsSingleton().Initialize(pd);
}

//============================  PRIVATE METHODS =================================/

//==============================================================================
// Set up the TCP Server on the device. Will set up a socket and listen
// for connection. On connection sends out welcome message and waits for msg.
//==============================================================================

void TcpServer::Initialize(const void * const pd)
{
	OSSemInit(&_msgProcessed, 0);
	OSMboxInit(&_mailbox, 0);

	//Set up the listening TCP socket
	int tcp_listen = listen(INADDR_ANY, TCP_LISTEN_PORT, 1); //max pending = 1 ; used to be 5

	if (tcp_listen > 0) //did the listen port setup
	{
		IPADDR client_addr;
		WORD port;

		while (true)
		{
			cout << "tcp:Waiting to accept a connection on port: " << TCP_LISTEN_PORT << "..." << endl;

			//Wait to accept an incoming connection
			_tcpConnection = accept(tcp_listen, &client_addr, &port, 0); //timeout =0 wait forever
			cout << "TCP: Connected to ";
			ShowIP(client_addr);
			cout << endl;
			SetConnectionStatus();
			ProcessMessagesWhileConnectionOpen(client_addr);
			cout << "Connection Ended" << endl;
			close(tcp_listen);
			tcp_listen = listen(INADDR_ANY, TCP_LISTEN_PORT, 1);
			cout << "TCP listen reset to: " << tcp_listen << endl;
		}//end while(True) listen port set up
	}
	else
	{
		cout << "ERROR: TCP Server not starting, can't set up listen port: " << TCP_LISTEN_PORT << endl;
		OSTimeDly(TICKS_PER_SECOND);
	}
}

void TcpServer::ProcessMessagesWhileConnectionOpen(IPADDR client_addr)
{
	while (_connectionIsOpen) //loop while connection is valid
	{
		int num_bytes_read = ReadUntilNoClientAvail(client_addr);

		if (num_bytes_read < 0)
		{
			CloseConnection();
		}
		else
		{
			ProcessMessage(num_bytes_read);
		}
	}
}


void TcpServer::ProcessMessage(int num_bytes_read)
{
	//There will always be room for the terminator because we only read a max of 1 less than full size
	_rxBuffer[num_bytes_read++] = '\0';
	if (_debuggingTcp) cout << "tcp:Read bytes:" << num_bytes_read << " string read = " << _rxBuffer << endl;
	_messageWaiting = true; //Set this first and process pending on mbox MUST call SetMessageProcessed();

	BYTE err_code = OSMboxPost(&_mailbox, static_cast<void*> (_rxBuffer)); //Put a message in the Mailbox.
	if ((err_code != OS_NO_ERR) && _debuggingTcp) cout << "OSMboxPost error in tcpServer "  << endl;
	//Wait for response from task that processes the message.
	err_code = OSSemPend(&_msgProcessed,MAX_WAIT_MSG_HANDLED_TICKS);
	if (err_code != OS_NO_ERR) cout << "OSSemPend Error in TcpServer "  <<endl;
}

//==============================================================================
// Read data from the client - if we don't get any within the timeout period
// see if a client is trying to reset the connection. If so return -1 which the
// caller will have to use to close the connection. Otherwise return
// the number of bytes read.
//==============================================================================
int TcpServer::ReadUntilNoClientAvail(IPADDR clientAddr)
{
	int num_bytes_read = 0;

	while (num_bytes_read == 0)
	{
		num_bytes_read = ReadWithTimeout(_tcpConnection, _rxBuffer, RX_BUF_SIZE - 1, READ_TIMEOUT_TICKS);
		if (num_bytes_read == 0)
		{
			//See if user is trying to reset the connection
			if (_tcpServerReset->IsConnectionReset())
			{
				_tcpServerReset->ClearConnectionReset();
				num_bytes_read = -1;
			}
		}
	} //end while num bytes read == 0
	return num_bytes_read;
}

void TcpServer::CloseConnection()
{
	close(_tcpConnection);
	_connectionIsOpen = false;
	_tcpConnection = 0; //no strictly necessary
	cout << "Ending connection -no response from client." << endl;
}

void TcpServer::SetConnectionStatus()
{
	if (_tcpConnection > 0) _connectionIsOpen = true;
	else _connectionIsOpen = false;
}

